/*
** Module   : WindowTitleBar.java
** Abstract : GUI window title bar implementation
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140929 Make class public.
** 006 CA  20141021 Finished support for the caption buttons, drawing, and associated system menu.
** 007 MAG 20140929 Eliminated call to AbstarctContainer.draw.
** 008 HC  20141021 Implemented window move/position support.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 CA  20141123 Fixed drawing of window title bar, to reflect if the window is the one with
**                  current focus or not.  Changes in drawing of the caption buttons/icon/title to
**                  reflect the OS drawing.
** 011 EVL 20141204 Adding title bar icon changing capability.
** 012 EVL 20150108 Taking into loaded icon size difference from either LOAD-SMALL-ICON or
**                  LOAD-ICON sources.  Fix the default application icon to be set as gclogo.ico
**                  embedded into runtime jar file.  Rework for icon/image implementation.
** 013 CA  20150323 Pass the title-bar's ID when constructing the window title, so the window can 
**                  be resolved.
** 014 CA  20150415 If the title is not set, use the default title.
** 015 HC  20150323 Removed business logic from config classes.
** 016 CA  20150406 Implemented nativeWidth and nativeHeight for WindowTitle class (change required
**                  after adding new Label APIs).
** 017 CA  20150705 Draw title centered vertically.
** 018 HC  20150510 Improved drawing of caption buttons - they are now more Windows alike. 
** 019 GES 20150425 Eliminated AWT dependencies.
** 020 HC  20150612 Implemented support for modal windows.
** 021 HC  20150702 Improved positioning of top-level windows on systems with multiple displays.
** 022 CA  20150826 Added mouse actions.
** 023 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
**                  Changes to allow multi-window focus management in GUI.
**                  Fix for title bar size and painting issues.
** 024 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 025 HC  20151021 Added the ability to enable or disable caption buttons.
** 026 EVL 20151125 Window activation fix. 
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;
import java.io.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.MouseEvt;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Implementation of a GUI window title bar.  This consists of (from left to right):
 * <ul>
 *    <li>an {@link #icon}</li>
 *    <li>a {@link #title}</li>
 *    <li>caption {@link #buttons}</li>
 * </ul>
 * Each element is a specific widget instance.
 */
public class WindowTitleBar
extends AbstractContainer<GuiOutputManager>
implements WidgetWithConfig<BaseConfig>,
           MouseListener,
           MouseMotionListener
{
   /** The window's icon. */
   WindowIcon icon;
   
   /** The window title. */
   WindowTitle title;

   /** The graphics driver. */
   private GuiDriver gd;

   /** The caption buttons. */
   private CaptionButton[] buttons;

   /** This widget configuration instance. */
   private BaseConfig config;
   
   /**
    * Basic c'tor.
    * 
    * @param    id
    *           The widget's ID.
    * @param    windowTitle
    *           The window title.
    */
   public WindowTitleBar(WidgetId id, String windowTitle)
   {
      this(id, true, new CaptionButtonType[]
         { 
            CaptionButtonType.ICONIFY, 
            CaptionButtonType.MAXIMIZE, 
            CaptionButtonType.CLOSE
         },
         windowTitle);
   }
   
   /**
    * Custom c'tor.
    * 
    * @param    id
    *           The widget's ID.
    * @param    showWindowIcon
    *           If <code>true</code> the window's window icon will be shown.
    * @param    buttonTypes
    *           Array of button types to create on the bar.
    * @param    windowTitle
    *           The window title.
    */
   public WindowTitleBar(WidgetId id, boolean showWindowIcon, CaptionButtonType[] buttonTypes,
                         String windowTitle)
   {
      nativeInsets = new NativeInsets(1, 2, 0, 2);

      gd = (GuiDriver) OutputManager.getDriver();
      
      config = ConfigManager.getInstance().resolveWidgetConfig(id, BaseConfig.class);
      
      if (showWindowIcon)
      {
         icon = new WindowIcon(WidgetId.nextID());
         
         byte[] appIcon = Window.getWindowDefaultIcon();
         
         // found the application icon defined on the server side
         if (appIcon != null)
         {
            // at this time we have either application icon or default icon loaded
            icon.setIconData(appIcon);
         }
            
         this.add(icon);
      }

      title = new WindowTitle(WidgetId.nextID());
      title.setTitle(windowTitle);
      this.add(title);

      buttons = new CaptionButton[buttonTypes.length];
      for (int i = 0; i < buttonTypes.length; i++)
      {
         CaptionButton btn = new CaptionButton(WidgetId.nextID(), buttonTypes[i]);
         this.add(btn);

         buttons[i] = btn;
      }
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
      // no-op
   }

   /**
    * Get the widget config.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Layout the elements of this title.   Everything is done in pixel units.
    */
   public void doLayout()
   {
      TopLevelWindow<GuiOutputManager> window = topLevelWindow();
      
      CoordinatesConversion c = screen().coordinates();

      int font = window.resolveTitleFont();
      // compute the window title's font height, in pixels.  this will determine the titlebar's
      // height.
      int fheight = c.heightToNative(1, window.getId(), font) + 
                    nativeInsets.top + 
                    nativeInsets.bottom;

      if (icon != null)
      {
         // - icon on left-side; to ensure the character unit location is properly in sync with the 
         // native location, compute the character location first and re-transform to native for the
         // physical location.
         icon.setLocation(c.widthFromNative(nativeInsets.left),
                          c.heightFromNative(nativeInsets.top));
         NativePoint iconLoc = c.pointToNative(icon.location());
         icon.setPhysicalLocation(iconLoc.x, iconLoc.y);
      }

      // title height can not be less than the default caption button height, 
      // also give 4 more pixels for margins
      fheight = Math.max(fheight, GuiConstants.CAPTION_BUTTON_HEIGHT + 4);
      
      // title height: needs to be done early, icon's height depends on it
      title.setHeight(c.heightFromNative(fheight));

      // entire bar height is compute in native units, then transformed to character units.
      // this is needed because when transforming metrics between native and character units,
      // these functions are not associative (i.e. T(a+b+c) does not necessarily give the same
      // result as T(a)+T(b)+T(c), where T is a transformation function).
      double barHeight = c.heightFromNative(c.heightToNative(title.height()) + 
                                            nativeInsets.top                 + 
                                            nativeInsets.bottom);
      setHeight(barHeight);
      
      // - minimize, maximize, close buttons
   
      double bw = c.widthFromNative(GuiConstants.CAPTION_BUTTON_WIDTH);
      double bh = c.heightFromNative(GuiConstants.CAPTION_BUTTON_HEIGHT);
      
      // add the buttons - all coordinates are relative to this container.
      double bx = this.width();
      // determine the vertical location so that the button is centered vertically
      double by = c.heightFromNative((c.heightToNative(barHeight) - GuiConstants.CAPTION_BUTTON_HEIGHT) / 2);

      for (int i = buttons.length - 1; i >= 0; i--)
      {
         CaptionButton btn  = buttons[i];
         btn.setDimension(bw, bh);
         
         int buttonSpace = GuiConstants.CAPTION_BUTTON_WIDTH + nativeInsets.right;
         
         if (btn.getType() == CaptionButtonType.ICONIFY)
         {
            // TODO: the buttons are touching each other except of the close button, 
            //       that has 2 pixels left-margin. 
            //       make this the button's property or at least make the algorithm more generic.
            buttonSpace -= 2;
         }
         
         bx = bx - c.widthFromNative(buttonSpace);
         
         btn.setLocation(bx, by);

         NativePoint btnLoc = c.pointToNative(btn.location());
         btn.setPhysicalLocation(btnLoc.x, btnLoc.y);
      }
      
      double iconColumn = 0;
      if (icon != null)
      {
         iconColumn = icon.bounds().right();
      }

      // - title in the middle, between the icon and the minimize button
      title.setLocation(iconColumn + 
                        c.widthFromNative(nativeInsets.left),
                        c.heightFromNative(nativeInsets.top));
      NativePoint titleLoc = c.pointToNative(title.location());
      title.setPhysicalLocation(titleLoc.x, titleLoc.y);

      // set title width
      double twidth = buttons[0].bounds().left() - 
                      title.location().x         - 
                      c.widthFromNative(nativeInsets.right);
      title.setWidth(twidth);
   }
   
   /**
    * Get widget minimum size - same as {@link #dimension()}.
    *
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }

   /**
    * Move this widget on the X and Y axis, by the specified deltas.  Values are specified in
    * units depending on the UI type.
    * 
    * @param   dx
    *          The X-axis delta.
    * @param   dy
    *          The Y-axis delta.
    */
   @Override
   public void moveBy(int dx, int dy)
   {
      topLevelWindow().moveBy(dx, dy);
   }

   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      Widget<GuiOutputManager> w = super.findMouseSource(p);
      
      return (w != null ? w : this);
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvt.MOUSE_MOVABLE
      };
   }
   
   /**
    * Draw the title's background.
    */
   @Override
   public void draw()
   {
      // drawing is done only once, after each (re)layout
      WindowTitleBar bar = WindowTitleBar.this;
      NativePoint origin = bar.physicalLocation();
      final NativeDimension dim = bar.physicalDimension();
      final NativeRectangle clip = clipRectangle(origin, dim);
      
      // fill title bar
      gd.draw(origin, clip, new Runnable()
      {
         @Override
         public void run()
         {
            int color = isActiveWindow() 
                           ? SysColor.COLOR_ACTIVECAPTION
                           : SysColor.COLOR_INACTIVECAPTION;
            
            gd.setColor(gd.getSysColor(color));
            gd.fillRect(0, 0, dim.width, dim.height - 1);

            // draw contained widgets
            for (Widget<?> widget : widgets())
            {
               widget.draw();
            }
         }
      });
   }
   
   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return false;
   }
   
   /**
    * Enabled or disables a caption button.
    * 
    * @param    button
    *           The button to enable/disable.
    * @param    enabled
    *           The new enabled state.
    */
   public void setCaptionButtonEnabled(CaptionButtonType button, boolean enabled)
   {
      for (CaptionButton b : buttons)
      {
         if (b.getType() == button)
         {
            b.setEnabled(enabled);
            break;
         }
      }
   }

   /**
    * Determine if the window associated with this bar is the active window.
    * 
    * @return   See above.
    */
   private boolean isActiveWindow()
   {
      TopLevelWindow<GuiOutputManager> w = topLevelWindow();
      WidgetId id = w.getId();
      // EVL<< DEBUGGING
      TopLevelWindow<GuiOutputManager> wATL = WindowManager.getActiveTopLevelWindow();
      WidgetId idATL = wATL.getId();
      System.out.println("Top Level ID: "+id.asInt()+" Active Top Level ID: "+idATL.asInt());
      
      TopLevelWindow<GuiOutputManager> aW = WindowManager.getActiveWindow();
      System.out.println("Active Window: "+(aW != null ? aW.getId().asInt() : aW));
      // EVL>> DEBUGGING
//      return WindowManager.getActiveTopLevelWindow().getId().equals(id);
      return aW != null && (aW.getId().equals(id) || !wATL.isVisible());
//      return WidgetId.DEFAULT_WINDOW_ID.equals(id) ? aW != null && aW.getId().equals(id)
//                : WindowManager.getActiveTopLevelWindow().getId().equals(id);
   }
   
   /**
    * The window's icon.
    */
   class WindowIcon
   extends ImageGuiImpl
   implements MouseListener
   {
      /** Flag indicating the icon was loaded by LOAD-SMALL-ICON(). */
      private boolean useSmallIconSize = false;
      
      /**
       * The boolean value indicates that the icon is changed. The flag is used
       * to track setIconData usages.
       */
      private boolean isIconChanged;

      /**
       * Create the window icon.
       * 
       * @param    id
       *           The widget's ID.
       */
      public WindowIcon(WidgetId id)
      {
         super(id, true);
      }
      
      /**
       * Get widget height.
       *
       * @return  widget height.
       */
      @Override
      public double height()
      {
         // 2 pixels should be reserved for icon margins
         return title.height() - screen().coordinates().heightFromNative(2);
      }

      /**
       * Get widget width.
       *
       * @return  widget width.
       */
      @Override
      public double width()
      {
         CoordinatesConversion c = screen().coordinates();

         return c.widthFromNative(c.heightToNative(height()));
      }

      /**
       * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
       * 
       * @return   Always <code>false</code>.
       */
      @Override
      public boolean raiseLegacyMouseEvents()
      {
         return false;
      }
      
      /**
       * Sets or replaces the window icon data with provided array data.
       * 
       * @param    newIcon
       *           A new icon image array data.
       *
       * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
       */
      public boolean setIconData(byte[] newIcon)
      {
         // the new image candidate
         ImageWrapper<?> imgNew = gd.createIcon(newIcon, useSmallIconSize);
         
         // image loaded, replace current with new one 
         if (imgNew != null)
         {
            img = imgNew;
            isIconChanged = true;
            
            // repaint changed icon
            repaint();

            return true;
         }
         else
         {
            // if error while data processing - do nothing and return false
            return false;
         }
      }

      /**
       * Loads the new application icon for given window from the client side filesystem.
       * 
       * @param    small
       *           Flag indicating usage of the small icon case(16x16 in size).
       * @param    name
       *           The fully qualified file name for the icon to be loaded.
       *
       * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
       */
      public boolean loadIcon(boolean small, String name)
      {
         InputStream is = null;
         byte[] icon = null;
         
         // find the icon on local file system
         try
         {
            is = new FileInputStream(name);
         }
         catch (FileNotFoundException ex)
         {
            is = null;
         }
         
         // the file is found, try to read
         if (is != null)
         {
            icon = UiUtils.getData(is);
         }
         
         // if having new icon data - set it
         if (icon != null)
         {
            useSmallIconSize = small;
            return setIconData(icon);
         }
         else
         {
            return false;
         }
      }

      /**
       * Sets the window icon, if the icon is changed and then draws it on the graphic device.
       */
      @Override
      public void draw()
      {
         if (isIconChanged && img != null)
         {
            isIconChanged = false;
            Object[] drawingParameters = calculateDrawingParameters();
            int iw = (Integer) drawingParameters[0];
            int ih = (Integer) drawingParameters[1];
            int offsetX = (Integer) drawingParameters[2];
            int offsetY = (Integer) drawingParameters[3];
            boolean transparent  = (Boolean) drawingParameters[4];
            boolean stretchToFit = (Boolean) drawingParameters[5];
            boolean retainShape  = (Boolean) drawingParameters[6];
            boolean convert3D    = (Boolean) drawingParameters[7];
            /**
             * Sets the window icon. The drawing parameters are due to the drawing operations are
             * performed on the server side for the web rendering engine implementation.
             */
            gd.setIcon(img, 0, 0, iw, ih, offsetX, offsetY, transparent,
                      stretchToFit, retainShape, convert3D);
         }

         super.draw();
      }
      
      
   }
   
   /**
    * The window's title.
    */
   private class WindowTitle
   extends Label<GuiOutputManager>
   {

      /** The cached window title */
      private String title;
      
      /** True value indicates that the title is changed and the graphics device must be notified.*/
      private boolean titleIsChanged;
      /**
       * Basic c'tor.
       * 
       * @param    id
       *           The widget's ID.
       */
      public WindowTitle(WidgetId id)
      {
         super(id, "", WindowTitleBar.this.getId().asInt());
      }

      /**
       * Returns the window title.
       * 
       * @return   The title
       */
      public String getTitle()
      {
         return title;
      }

      /**
       * Sets the window title.
       * 
       * @param    title
       *           The title to set
       */
      public void setTitle(String title)
      {
         this.title = title;
         this.titleIsChanged = true;
      }

      /**
       * Draw this window title.
       */
      @Override
      public void draw()
      {
         final TopLevelWindow<GuiOutputManager> window = topLevelWindow();
         String windowTitle = window.title();

         final String title = (windowTitle == null ? WindowGuiImpl.DEFAULT_TITLE : windowTitle);
         
         if (titleIsChanged || !title.equals(getTitle()))
         {
            // notify client about new title
            gd.setTitle(title);
            titleIsChanged = false;
         }
         NativePoint origin = physicalLocation();
         final NativeDimension d = physicalDimension();
         final NativeRectangle clip = clipRectangle(origin, d);
         
         final int titleFont = window.resolveTitleFont();
         config.font = titleFont;
         final GuiFontResolver gf = new GuiFontResolver(window.window(), gd, config);
         
         gd.draw(origin, clip, new Runnable()
         {
            @Override
            public void run()
            {
               // EVL<< DEBUGGING
               System.out.println("Draw title for Window: "+title);
               // EVL>> DEBUGGING
               int color = isActiveWindow()
                              ? SysColor.COLOR_ACTIVECAPTION
                              : SysColor.COLOR_INACTIVECAPTION;

               // remove old title
               gd.setColor(gd.getSysColor(color));
               gd.fillRect(0, 0, d.width, d.height - 1);

               if (title.length() > 0)
               {
                  FontDetails<?> font = gf.font().font;

                  // draw text
                  gd.setColor(gd.getSysColor(SysColor.COLOR_CAPTIONTEXT));
                  gd.setGuiFont(font);

                  gd.drawStringCentered(title, 0, d.height);
               }
            }
         });
      }

      /**
       * Draw caret - this is a no-op.
       */
      @Override
      public void drawCaret()
      {
         // no-op
      }
      
      /**
       * Find the widget to which the given mouse event can be applied.
       * 
       * @param    p
       *           The mouse physical location.
       * 
       * @return   A widget over which the mouse is positioned or the associated 
       *           {@link WindowTitleBar} if none found.
       */
      @Override
      public Widget<GuiOutputManager> findMouseSource(NativePoint p)
      {
         Widget<GuiOutputManager> w = super.findMouseSource(p);
         
         return (w == null ? WindowTitleBar.this : w);
      }
      
      /**
       * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
       * 
       * @return   Always <code>false</code>.
       */
      @Override
      public boolean raiseLegacyMouseEvents()
      {
         return false;
      }
      
      /**
       * Retrieve the widget's width in native units.
       * 
       * @return   See above.
       */
      @Override
      protected int nativeWidth()
      {
         return screen().coordinates().widthToNative(width());
      }
      
      /**
       * Retrieve the widget's height in native units.
       * 
       * @return   See above.
       */
      @Override
      protected int nativeHeight()
      {
         return screen().coordinates().heightToNative(height());
      }
   }
}
